@@ -1,5 +1,6 @@
-# Copyright (C) 2004-2019 Aaron Swartz
+# Copyright (C) 2004-2020 Aaron Swartz
 #                         Andrey Zelenchuk <azelenchuk@parallels.com>
+#                         Andrey Zelenchuk <azelenchuk@plesk.com>
 #                         Brian Lalor
 #                         Dean Jackson
 #                         Erik Hetzner
@@ -7,6 +8,7 @@
 #                         Joey Hess
 #                         Kaashif Hymabaccus <kaashif@kaashif.co.uk>
 #                         Lindsey Smith <lindsey.smith@gmail.com>
+#                         Léo Gaspard <leo@gaspard.io>
 #                         Marcel Ackermann
 #                         Martin 'Joey' Schulze
 #                         Matej Cepl
@@ -43,15 +45,11 @@ from . import config as _config
 from . import error as _error
 from . import feed as _feed
 
-UNIX = False
 try:
     import fcntl as _fcntl
-    # A pox on SunOS file locking methods
-    if 'sunos' not in _sys.platform:
-        UNIX = True
-except:
-    pass
-
+    UNIX = True
+except ImportError:
+    UNIX = False
 
 # Path to the filesystem root, '/' on POSIX.1 (IEEE Std 1003.1-2008).
 ROOT_PATH = _os.path.splitdrive(_sys.executable)[0] or _os.sep
@@ -132,18 +130,18 @@ class Feeds (list):
     datafile_version = 2
     datafile_encoding = 'utf-8'
 
-    def __init__(self, configfiles=None, datafile=None, config=None):
+    def __init__(self, configfiles=None, datafile_path=None, config=None):
         super(Feeds, self).__init__()
         if configfiles is None:
             configfiles = self._get_configfiles()
         self.configfiles = configfiles
-        if datafile is None:
-            datafile = self._get_datafile()
-        self.datafile = _os.path.realpath(datafile)
+        if datafile_path is None:
+            datafile_path = self._get_datafile_path()
+        self.datafile_path = _os.path.realpath(datafile_path)
         if config is None:
             config = _config.CONFIG
         self.config = config
-        self._datafile_lock = None
+        self.datafile = None
 
     def __getitem__(self, key):
         for feed in self:
@@ -214,7 +212,7 @@ class Feeds (list):
                 [_os.path.join(config_dir, 'rss2email.cfg')
                  for config_dir in config_dirs]))
 
-    def _get_datafile(self):
+    def _get_datafile_path(self):
         """Get the data file path
 
         Following the XDG Base Directory Specification.
@@ -238,7 +236,7 @@ class Feeds (list):
                 return datafile
         return datafiles[0]
 
-    def load(self, lock=True, require=False):
+    def load(self, require=False):
         _LOG.debug('load feed configuration from {}'.format(self.configfiles))
         if self.configfiles:
             self.read_configfiles = self.config.read(self.configfiles)
@@ -246,30 +244,28 @@ class Feeds (list):
             self.read_configfiles = []
         _LOG.debug('loaded configuration from {}'.format(
                 self.read_configfiles))
-        self._load_feeds(lock=lock, require=require)
+        self._load_feeds(require=require)
 
-    def _load_feeds(self, lock, require):
-        _LOG.debug('load feed data from {}'.format(self.datafile))
-        if not _os.path.exists(self.datafile):
+    def _load_feeds(self, require):
+        _LOG.debug('load feed data from {}'.format(self.datafile_path))
+        if not _os.path.exists(self.datafile_path):
             if require:
                 raise _error.NoDataFile(feeds=self)
-            _LOG.info('feed data file not found at {}'.format(self.datafile))
+            _LOG.info('feed data file not found at {}'.format(self.datafile_path))
             _LOG.debug('creating an empty data file')
-            dirname = _os.path.dirname(self.datafile)
+            dirname = _os.path.dirname(self.datafile_path)
             if dirname and not _os.path.isdir(dirname):
                 _os.makedirs(dirname, mode=0o700, exist_ok=True)
-            with _codecs.open(self.datafile, 'w', self.datafile_encoding) as f:
+            with _codecs.open(self.datafile_path, 'w', self.datafile_encoding) as f:
                 self._save_feed_states(feeds=[], stream=f)
         try:
-            self._datafile_lock = _codecs.open(
-                self.datafile, 'r', self.datafile_encoding)
+            self.datafile = _codecs.open(
+                self.datafile_path, 'r', self.datafile_encoding)
         except IOError as e:
             raise _error.DataFileError(feeds=self) from e
 
-        locktype = 0
-        if lock and UNIX:
-            locktype = _fcntl.LOCK_SH
-            _fcntl.lockf(self._datafile_lock.fileno(), locktype)
+        if UNIX:
+            _fcntl.lockf(self.datafile, _fcntl.LOCK_SH)
 
         self.clear()
 
@@ -277,10 +273,10 @@ class Feeds (list):
         handlers = list(_LOG.handlers)
         feeds = []
         try:
-            data = _json.load(self._datafile_lock)
+            data = _json.load(self.datafile)
         except ValueError as e:
             _LOG.info('could not load data file using JSON')
-            data = self._load_pickled_data(self._datafile_lock)
+            data = self._load_pickled_data(self.datafile)
         version = data.get('version', None)
         if version != self.datafile_version:
             data = self._upgrade_state_data(data)
@@ -291,16 +287,12 @@ class Feeds (list):
                 raise _error.DataFileError(
                     feeds=self,
                     message='missing feed name in datafile {}'.format(
-                        self.datafile))
+                        self.datafile_path))
             feeds.append(feed)
         _LOG.setLevel(level)
         _LOG.handlers = handlers
         self.extend(feeds)
 
-        if locktype == 0:
-            self._datafile_lock.close()
-            self._datafile_lock = None
-
         for feed in self:
             feed.load_from_config(self.config)
 
@@ -322,7 +314,7 @@ class Feeds (list):
 
     def _load_pickled_data(self, stream):
         _LOG.info('try and load data file using Pickle')
-        with open(self.datafile, 'rb') as f:
+        with open(self.datafile_path, 'rb') as f:
             feeds = list(feed.get_state() for feed in _pickle.load(f))
         return {
             'version': self.datafile_version,
@@ -358,19 +350,27 @@ class Feeds (list):
         self._save_feeds()
 
     def _save_feeds(self):
-        _LOG.debug('save feed data to {}'.format(self.datafile))
-        dirname = _os.path.dirname(self.datafile)
+        _LOG.debug('save feed data to {}'.format(self.datafile_path))
+        dirname = _os.path.dirname(self.datafile_path)
         if dirname and not _os.path.isdir(dirname):
             _os.makedirs(dirname, mode=0o700, exist_ok=True)
-        tmpfile = self.datafile + '.tmp'
+        tmpfile = self.datafile_path + '.tmp'
         with _codecs.open(tmpfile, 'w', self.datafile_encoding) as f:
             self._save_feed_states(feeds=self, stream=f)
             f.flush()
             _os.fsync(f.fileno())
-        _os.replace(tmpfile, self.datafile)
-        if UNIX and self._datafile_lock is not None:
-            self._datafile_lock.close()  # release the lock
-            self._datafile_lock = None
+        if UNIX:
+            # Replace the file, then release the lock by closing the old one.
+            _os.replace(tmpfile, self.datafile_path)
+            if self.datafile is not None:
+                self.datafile.close()  # release the lock
+                self.datafile = None
+        else:
+            # On Windows we cannot replace the file while it is opened. And we have no lock.
+            if self.datafile is not None:
+                self.datafile.close()
+                self.datafile = None
+            _os.replace(tmpfile, self.datafile_path)
 
     def _save_feed_states(self, feeds, stream):
         _json.dump(
